from typing import Any, Optional

from . import Payload, PayloadCollection, Prompter, PrompterHub, UserPayload
from .general import (CallablePrompter, ConsolePrompter,
                      TkinterMessageBoxPrompter)
from .hubs import PrompterBroadcaster, PrompterSwitcher


class PrompterBuilder:
    """Builder for prompter."""

    def __init__(self) -> None:
        self._result: Optional[Prompter] = None

    def use(self, prompter: Prompter) -> "PrompterBuilder":
        """Use a prompter."""
        if self._result is None:
            self._result = prompter
        elif isinstance(self._result, PrompterHub):
            self._result.register(prompter)
        else:
            raise Exception(
                f"The prompter is set to a no-hub prompter {type(self._result)}.")
        return self

    def useBroadcaster(self, final: bool = False) -> "PrompterBuilder":
        """Use broadcaster as hub."""
        return self.use(PrompterBroadcaster(final))

    def useSwitcher(self, final: bool = False) -> "PrompterBuilder":
        """Use switcher as hub."""
        return self.use(PrompterSwitcher(final))

    def useConsole(self, final: bool = False) -> "PrompterBuilder":
        """Use console prompter."""
        return self.use(ConsolePrompter(final))

    def useCallable(self, final: bool = False) -> "PrompterBuilder":
        """Use callable prompter."""
        return self.use(CallablePrompter(final))

    def useTkinterMessageBox(self, final: bool = False) -> "PrompterBuilder":
        """Use Tkinter message box prompter."""
        return self.use(TkinterMessageBoxPrompter(final))

    def clear(self) -> "PrompterBuilder":
        """Clear all prompters."""
        self._result = None
        return self

    def build(self) -> Optional[Prompter]:
        """Build the final prompter."""
        return self._result


class PayloadBuilder:
    """Builder for payload."""

    def __init__(self) -> None:
        self._result: Optional[Payload] = None

    def use(self, payload: Any) -> "PayloadBuilder":
        """Use a payload."""
        wrapped = payload if isinstance(
            payload, Payload) else UserPayload(payload)
        if self._result is None:
            self._result = wrapped
        elif isinstance(self._result, PayloadCollection):
            self._result.withPayload(wrapped)
        else:
            payloads = PayloadCollection([self._result, wrapped])
            self._result = payloads
        return self

    def build(self) -> Optional[Payload]:
        """Build the final payload."""
        return self._result



from enum import auto
from typing import Any

import click

from . import PayloadCollection, Prompter, PromptResult


class TkinterMessageBoxPrompter(Prompter):
    def __init__(self, final: bool = False) -> None:
        super().__init__(final)

    def prompt(self, payloads: PayloadCollection) -> PromptResult:
        import tkinter
        import tkinter.messagebox
        from ..helpers import buildMessage

        top = tkinter.Tk()
        top.withdraw()

        tkinter.messagebox.showinfo(
            f"Attention {buildMessage(payloads)}", str(payloads))
        top.destroy()

        return self.success()


class ConsolePrompter(Prompter):
    def __init__(self, final: bool = False) -> None:
        super().__init__(final)

    def prompt(self, payloads: PayloadCollection) -> PromptResult:
        from ..helpers import buildMessage

        click.echo(f"Attention {buildMessage(payloads)}: {payloads}")

        return self.success()


class CallablePrompter(Prompter):
    def __init__(self, final: bool = False) -> None:
        super().__init__(final)

    def prompt(self, payloads: PayloadCollection) -> PromptResult:
        hasCallable = False
        for payload in payloads.getUsers():
            if callable(payload.payload):
                payload.payload()
                hasCallable = True
        return self.success() if hasCallable else PromptResult.Unsupported



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Iterable, Iterator, List, Optional, Union, cast


class PromptResult(Enum):
    """Result for prompting."""
    Empty = 0
    Unsupported = 1     # Unsupported, skip this and go next
    Resolved = 2        # Resolved, go next
    Finished = 3        # Resolved, stop going next
    Failed = 4          # Failed, stop going next


class Payload(ABC):
    """Payload for prompters."""
    pass


@dataclass
class CyclePayload(Payload):
    """Payload for cycle event."""
    work: bool
    index: int


@dataclass
class UserPayload(Payload):
    """Payload that user provided."""
    payload: Any


@dataclass
class SchedulePayload(Payload):
    """Payload for scheduler information."""
    index: int
    message: str
    startTime: datetime
    endTime: datetime
    duration: timedelta = field(init=False)

    def __post_init__(self) -> None:
        self.duration = self.endTime - self.startTime


class PayloadCollection(Payload):
    """A collection for multiple payloads."""

    def __init__(self, payloads: Optional[List[Payload]] = None) -> None:
        super().__init__()
        self.payloads = payloads if payloads is not None else []

    def withPayload(self, payload: Payload) -> "PayloadCollection":
        self.payloads.append(payload)
        return self

    def withPayloads(self, payloads: "PayloadCollection") -> "PayloadCollection":
        self.payloads.extend(payloads)
        return self

    def tryGet(self, type: type) -> Iterable[Payload]:
        for item in self.payloads:
            if isinstance(item, type):
                yield item

    def getSchedule(self) -> Optional[SchedulePayload]:
        return next(iter(self.tryGet(SchedulePayload)), None)

    def getCycle(self) -> Optional[CyclePayload]:
        return next(iter(self.tryGet(CyclePayload)), None)

    def getUsers(self) -> Iterable[UserPayload]:
        return self.tryGet(UserPayload)

    def __len__(self) -> int:
        return len(self.payloads)

    def __getitem__(self, index) -> Payload:
        return self.payloads[index]

    def __repr__(self) -> str:
        return f"PayloadCollection({len(self.payloads)} payloads: {'; '.join([str(x) for x in self.payloads])})"

    def __str__(self) -> str:
        return f"Payloads({len(self.payloads)}: {'; '.join([type(x).__name__ for x in self.payloads])})"


class Prompter(ABC):
    """Prompter for every event."""

    def __init__(self, final: bool = False) -> None:
        self.final = final

    def success(self) -> PromptResult:
        return PromptResult.Finished if self.final else PromptResult.Resolved

    @abstractmethod
    def prompt(self, payloads: PayloadCollection) -> PromptResult:
        pass

    def __repr__(self) -> str:
        return type(self).__name__


class PrompterHub(Prompter, ABC):
    """Hub for multiple prompters."""
    @abstractmethod
    def register(self, prompter: Prompter) -> None:
        pass



import functools
import logging
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from typing import Any, Optional, Union

from ..prompters import Prompter
from ..timeutils import parse_time, subtract_time, to_timedelta


@dataclass(order=True, frozen=True)
class TimeTableItem:
    time: time
    message: str = ""
    payload: Any = None
    cycleIndex: Optional[int] = None
    cycleWork: bool = False


class TimeTable:
    _logger = logging.getLogger("TimeTable")

    def __init__(self) -> None:
        self.items: list[TimeTableItem] = []
        self.prompter: Optional[Prompter] = None

    def use(self, prompter: Optional[Prompter]) -> None:
        self._logger.debug(f"Use prompter {prompter}.")
        self.prompter = prompter

    def at(self, time: time, message: str = "", payload: Any = None, cycleIndex: Optional[int] = None, cycleWork: bool = False) -> None:
        self._logger.debug(f"{message} ({payload}) at {time}.")
        self.items.append(TimeTableItem(
            time, message, payload, cycleIndex, cycleWork))

    def cycle(self, start: time, end: time, work_duration: timedelta, rest_duration: timedelta, message: str = "", work_payload: Any = None, rest_payload: Any = None) -> None:
        self._logger.debug(
            f"{message} ({work_payload}, {rest_payload}) cycle from {start} to {end} (work {work_duration}, rest {rest_duration}).")
        _start = datetime(2000, 1, 1) + to_timedelta(start)
        _end = datetime(2000, 1, 1) + to_timedelta(end)

        index = 0

        current = _start

        while current < _end:
            index += 1
            self.at(min(current, _end).time(),
                    message, work_payload, index, True)
            current += work_duration
            self.at(min(current, _end).time(),
                    message, rest_payload, index, False)
            current += rest_duration

    def clear(self) -> None:
        self.items.clear()



from datetime import datetime, time, timedelta

BASE_DATETIME = datetime(2000, 1, 1)


def to_timedelta(value: time) -> timedelta:
    return timedelta(hours=value.hour, minutes=value.minute,
                     seconds=value.second, microseconds=value.microsecond)


def subtract_time(a: time, b: time) -> timedelta:
    """Return the time delta (a - b)."""
    return to_timedelta(a) - to_timedelta(b)


def parse_time(value: str) -> time:
    """Parse 00:00 or 00:00:00 to time."""
    return time(*list(map(int, map(round, map(float, value.split(':'))))))


def time_to_today(value: time) -> datetime:
    """Transform time to today time."""
    now = datetime.now()
    return datetime(year=now.year, month=now.month, day=now.day, tzinfo=now.tzinfo) + to_timedelta(value)



import functools
import json
import logging
from datetime import date, datetime, time, timedelta
from typing import Any, Dict, Optional, Union, cast

import click

import enlighten

from ..extensions import (findExtensions, loadExtension, loadExtensions,
                          useExtension, useExtensions)
from ..prompters import Prompter
from ..prompters.builders import PayloadBuilder, PrompterBuilder
from ..timeutils import parse_time, subtract_time, to_timedelta
from .timetable import TimeTable, TimeTableItem


def default_prompter_builder() -> PrompterBuilder:
    prompter = PrompterBuilder()
    prompter.useSwitcher().useConsole().useCallable(True).useTkinterMessageBox()
    return prompter


class SchemaBuilder:
    _logger = logging.getLogger("SchemaBuilder")

    def __init__(self) -> None:
        self.result: TimeTable = TimeTable()

    def getEnv(self) -> Dict[str, Any]:
        prompterBuilder = default_prompter_builder()

        env = {}

        def at(raw_time: Union[str, time], message: str = "", payload: Any = None) -> None:
            ttime = parse_time(raw_time) if isinstance(
                raw_time, str) else raw_time
            if isinstance(payload, PayloadBuilder):
                payload = payload.build()
            self.result.at(ttime, message, payload)

        def cycle(raw_start: Union[str, time], raw_end: Union[str, time], raw_work_duration: Union[str, time, timedelta], raw_rest_duration: Union[str, time, timedelta], message: str = "", work_payload: Any = None, rest_payload: Any = None) -> None:
            tstart = parse_time(raw_start) if isinstance(
                raw_start, str) else raw_start
            tend = parse_time(raw_end) if isinstance(raw_end, str) else raw_end
            twork_duration = to_timedelta(parse_time(raw_work_duration)) if isinstance(
                raw_work_duration, str) else to_timedelta(raw_work_duration) if isinstance(
                raw_work_duration, time) else raw_work_duration
            trest_duration = to_timedelta(parse_time(raw_rest_duration)) if isinstance(
                raw_rest_duration, str) else to_timedelta(raw_rest_duration) if isinstance(
                raw_rest_duration, time) else raw_rest_duration

            if isinstance(work_payload, PayloadBuilder):
                work_payload = work_payload.build()
            if isinstance(rest_payload, PayloadBuilder):
                rest_payload = rest_payload.build()

            self.result.cycle(
                tstart, tend, twork_duration, trest_duration,
                message, work_payload, rest_payload)

        def payloads() -> PayloadBuilder:
            return PayloadBuilder()

        def loadRaw(source: str) -> None:
            src_preview = source[:50].replace('\n', ' ').replace('\r', ' ')
            self._logger.info(f"Load: '{src_preview}...'")
            exec(source, env)

        def load(file: str, encoding: str = "utf8") -> None:
            with open(file, encoding=encoding) as f:
                source = f.read()
            loadRaw(source)

        def ext(name: Optional[str] = None) -> None:
            if name is None:
                self._logger.info("Use all installed extensions.")
                extnames = findExtensions()
                exts = loadExtensions(extnames)
                useExtensions(exts, env)
            else:
                self._logger.info(f"Use extension {name}.")
                extension = loadExtension(name)
                useExtension(extension, env)

        env["at"] = at
        env["cycle"] = cycle
        env["load"] = load
        env["loadRaw"] = loadRaw
        env["ext"] = ext
        env["payloads"] = payloads
        env["prompter"] = prompterBuilder
        env["env"] = env

        return env

    def useEnv(self, env: Dict[str, Any]):
        prompter = env.get("prompter")
        if isinstance(prompter, PrompterBuilder):
            self.result.use(prompter.build())

    def loadWithEnv(self, src: str, env: Dict[str, Any]) -> None:
        env["loadRaw"](src)
        self.useEnv(env)

    def load(self, src) -> None:
        self.loadWithEnv(src, self.getEnv())




import functools
import json
import logging
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from time import sleep
from typing import Any, Iterable, List, Optional, Union

import click

import enlighten

from ..prompters import PayloadCollection
from ..schedulers import ScheduledTimeTableItem

EMOJI_TIME = ["🕛🕐🕑🕒🕓🕔🕕🕖🕗🕘🕙🕚", "🕧🕜🕝🕞🕟🕠🕡🕢🕣🕤🕥🕦"]
EMOJI_BELL = "🔔"
EMOJI_WORK = "💼"
EMOJI_REST = "☕"


def buildMessage(payloads: Union[PayloadCollection, ScheduledTimeTableItem], cycle: bool = True, icon: bool = True) -> str:
    if isinstance(payloads, ScheduledTimeTableItem):
        return buildMessage(payloads.buildPayloads(), cycle=cycle, icon=icon)

    schedule = payloads.getSchedule()

    iconStartTime = EMOJI_TIME[0 if schedule.startTime.minute <
                               30 else 1][schedule.startTime.hour % 12]
    iconEndTime = EMOJI_TIME[0 if schedule.endTime.minute <
                             30 else 1][schedule.endTime.hour % 12]
    iconHead = EMOJI_BELL

    schedule = payloads.getSchedule()
    message = schedule.message

    if cycle:
        cycleP = payloads.getCycle()
        if cycleP:
            iconHead = EMOJI_WORK if cycleP.work else EMOJI_REST
            message += f" (cycle {cycleP.index} {'' if cycleP.work else 'resting '}starting)"

    if icon:
        return f"{iconStartTime} {schedule.startTime.time()} - {iconEndTime} {schedule.endTime.time()} {iconHead} {message}"
    else:
        return f"{schedule.startTime.time()} - {schedule.endTime.time()} {message}"



import functools
import json
import logging
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from time import sleep
from typing import Any, Iterable, List, Optional, Union

import click

import enlighten

from ..prompters import (CyclePayload, Payload, PayloadCollection, Prompter,
                         SchedulePayload, UserPayload)
from ..schemas.timetable import TimeTable, TimeTableItem
from ..timeutils import subtract_time, time_to_today


@dataclass(frozen=True)
class ScheduledTimeTableItem:
    raw: TimeTableItem
    index: int
    startTime: datetime
    endTime: datetime

    def buildPayloads(self) -> PayloadCollection:
        payloads = PayloadCollection().withPayload(
            SchedulePayload(self.index, self.raw.message, self.startTime, self.endTime))

        if self.raw.cycleIndex is not None:
            payloads.withPayload(CyclePayload(
                self.raw.cycleWork, self.raw.cycleIndex))

        if isinstance(self.raw.payload, PayloadCollection):
            for payload in self.raw.payload:
                if payload is not None:
                    payloads.withPayload(payload)
        elif isinstance(self.raw.payload, Payload):
            payloads.withPayload(self.raw.payload)
        elif self.raw.payload is not None:
            payloads.withPayload(UserPayload(self.raw.payload))

        return payloads


class Scheduler:
    _logger = logging.getLogger("Scheduler")

    def __init__(self, enableIcon: bool = True) -> None:
        self._STR_PENDING = "⏳" if enableIcon else "Pending:"
        self._STR_ATTENTION = "📣" if enableIcon else "Attention:"
        self._STR_OUTDATED = "📌" if enableIcon else "Outdated:"

    def scheduledTimeTable(self, timetable: TimeTable) -> Iterable[ScheduledTimeTableItem]:
        items: List[TimeTableItem] = list(sorted(timetable.items))
        totalLen = len(items)
        for index, item in enumerate(items):
            nextItem = None if index + \
                1 >= totalLen else items[index + 1]
            yield ScheduledTimeTableItem(item, index, time_to_today(
                item.time), time_to_today(nextItem.time if nextItem else item.time))

    def schedule(self, timetable: TimeTable, prompter: Optional[Prompter] = None) -> None:
        from ..helpers import buildMessage

        def outdating(item: ScheduledTimeTableItem) -> bool:
            now = datetime.now().time()

            raw = item.raw

            if raw.time < now:
                self._logger.info(f"Outdated: {item}.")
                click.echo(f"{self._STR_OUTDATED} {buildMessage(item)}")
                return True

            return False

        def pending(item: ScheduledTimeTableItem, status: enlighten.StatusBar, manager: enlighten.Manager) -> bool:
            now = datetime.now().time()

            raw = item.raw

            if raw.time <= now:
                return False

            self._logger.info(f"Pending: {raw}.")

            message = buildMessage(item)

            status.update(f"{self._STR_PENDING} {message}")
            deltaNow = subtract_time(raw.time, now)
            pendingTotal = int(round(deltaNow.total_seconds()))

            with manager.counter(
                    total=pendingTotal, desc="", unit='ticks', leave=False) as pbar:

                while True:
                    now = datetime.now().time()
                    if raw.time <= now:
                        self._logger.info(f"Occurring: {raw}.")
                        click.echo(f"{self._STR_ATTENTION} {message}")

                        result = prompter.prompt(item.buildPayloads())
                        self._logger.info(f"Prompting result: {result}.")
                        return True
                    else:
                        delta = subtract_time(raw.time, now)
                        count = int(round(delta.total_seconds()))
                        pbar.update((pbar.total - count) - pbar.count)
                    sleep(1)

            return False

        self._logger.info(f"Start scheduling.")
        click.echo(f"Started Time: {datetime.now().time()}")

        prompter = timetable.prompter if prompter is None else prompter

        if prompter is None:
            from ..schemas import default_prompter_builder
            prompter = default_prompter_builder().build()

        self._logger.info(f"Used prompter: {prompter}.")

        with enlighten.get_manager() as manager:
            with manager.status_bar('Status',
                                    color='white_on_cyan',
                                    justify=enlighten.Justify.CENTER, leave=False) as status:

                for scheduled in self.scheduledTimeTable(timetable):
                    while True:
                        if outdating(scheduled):
                            break
                        elif pending(scheduled, status, manager):
                            break
